AUTH_HOST = "https://www.googleapis.com"
AUTH_VERIFY_URL = "/identitytoolkit/v3/relyingparty/verifyPassword"
AUTH_ACCT_INFO_URL = "/identitytoolkit/v3/relyingparty/getAccountInfo"
AUTH_TIMEOUT_SECONDS = 15

AUTH_REFRESH_ENDPOINT = "https://securetoken.googleapis.com/v1/token"

FIREBASE_APP_KEY = "AIzaSyBi1DFJlBr9Cezf2BwfaT-PRPYmi3X3pdA"
FIREBASE_DEFAULT_URL = "https://elife-prod.firebaseio.com"
FIREBASE_INSTALLATIONS_PATH = "/installations2.json"
FIREBASE_DEVICES_PATH_BY_ID = "/devices/{}.json"
FIREBASE_DEVICE_DATA_PATH_BY_ID = "/devices/{}/data.json"



import datetime as dt

DEFAULT_TIME_ZONE: dt.tzinfo = dt.timezone.utc


def now(time_zone=None) -> dt.datetime:
    """Get now in specified time zone."""
    return dt.datetime.now(time_zone or DEFAULT_TIME_ZONE)



"""Rointe API Client"""
from __future__ import annotations
import logging
from typing import Any, Dict, Optional
import requests
from collections import namedtuple
from datetime import datetime, time, timedelta
from rointesdk.device import RointeDevice, ScheduleMode

from rointesdk.settings import (
    AUTH_ACCT_INFO_URL,
    AUTH_HOST,
    AUTH_REFRESH_ENDPOINT,
    AUTH_TIMEOUT_SECONDS,
    AUTH_VERIFY_URL,
    FIREBASE_APP_KEY,
    FIREBASE_DEFAULT_URL,
    FIREBASE_DEVICE_DATA_PATH_BY_ID,
    FIREBASE_DEVICES_PATH_BY_ID,
    FIREBASE_INSTALLATIONS_PATH,
)

_LOGGER = logging.getLogger(__name__)

ApiResponse = namedtuple("ApiResponse", ["success", "data", "error_message"])


class RointeAPI:
    def __init__(self, username: str, password: str):

        self.username = username
        self.password = password

        self.refresh_token = None
        self.auth_token = None
        self.auth_token_expire_date = None

        self._initialize_authentication()

    def _initialize_authentication(self) -> None:
        """Initializes the refresh token and cleans
            the original credentials."""

        login_data = self.login_user(self.username, self.password)

        if not login_data.success:
            _LOGGER.error(
                "Unable to authenticate user: %s",
                login_data.error_message)

            self.auth_token = None
            self.refresh_token = None
            return

        self.auth_token = login_data["auth_token"]
        self.refresh_token = login_data["refresh_token"]
        self.auth_token_expire_date = login_data["expires"]

        self._clean_credentials()

    def _clean_credentials(self):
        """Cleans authentication values"""
        self.username = None
        self.password = None

    def _ensure_valid_auth(self) -> bool:
        """Ensure there is a valid authentication token present."""

        now = datetime.now()

        if not self.auth_token or (
            self.auth_token_expire_date and self.auth_token_expire_date < now
        ):
            if not self._refresh_token():
                return False

        return True

    def _refresh_token(self) -> bool:
        """Refreshes authentication."""

        payload = {"grant_type": "refresh_token", "refresh_token": self.refresh_token}

        response = requests.post(
            f"{AUTH_REFRESH_ENDPOINT}?key={FIREBASE_APP_KEY}",
            data=payload,
            timeout=AUTH_TIMEOUT_SECONDS,
        )

        if not response:
            _LOGGER.error("No response while refreshing authentication")
            return False

        if response.status_code != 200:
            _LOGGER.error(
                "Invalid response [%s] when refreshing authentication",
                response.status_code,
            )
            return False

        response_json = response.json()

        if not response_json or "idToken" not in response_json:
            _LOGGER.error("Error while refreshing authentication")
            return False

        self.auth_token = response_json["idToken"]
        self.auth_token_expire_date = datetime.now() + timedelta(
            seconds=int(response_json["expiresIn"])
        )
        self.refresh_token = response_json["refresh_token"]

        return True

    def login_user(self, username: str, password: str) -> ApiResponse:
        """Log the user in."""

        payload = {"email": username, "password": password, "returnSecureToken": True}

        response = requests.post(
            f"{AUTH_HOST}{AUTH_VERIFY_URL}?key={FIREBASE_APP_KEY}",
            data=payload,
            timeout=AUTH_TIMEOUT_SECONDS,
        )

        if not response:
            return ApiResponse(False, None, "No response while authenticating")

        if response.status_code != 200:
            return ApiResponse(
                False,
                None,
                f"Invalid response code {response.status_code} while authenticating",
            )

        response_json = response.json()

        if not response_json or "idToken" not in response_json:
            return ApiResponse(False, None, "API Error while logging in")

        data = {
            "auth_token": response_json["idToken"],
            "expires": datetime.now()
            + timedelta(seconds=int(response_json["expiresIn"])),
            "refresh_token": response_json["refresh_token"],
        }

        return ApiResponse(True, data, None)

    def get_local_id(self, auth_token: str) -> ApiResponse:
        """Retrieve user local_id value."""

        if not auth_token:
            return ApiResponse(False, None, "Authentication not present")

        payload = {"idToken": auth_token}

        response = requests.post(
            f"{AUTH_HOST}{AUTH_ACCT_INFO_URL}?key={FIREBASE_APP_KEY}",
            data=payload,
        )

        if not response:
            return ApiResponse(
                False, None, "No response from API call to get_local_id()"
            )

        if response.status_code != 200:
            return ApiResponse(
                False, None, f"get_local_id() returned {response.status_code}"
            )

        response_json = response.json()

        return ApiResponse(True, response_json["users"][0]["localId"], None)

    def get_installation_by_id(
        self, installation_id: str, local_id: str, auth_token: str
    ) -> ApiResponse:
        """Retrieve a specific installation by ID."""

        args = {"auth": auth_token, "orderBy": '"userid"', "equalTo": f'"{local_id}"'}

        url = f"{FIREBASE_DEFAULT_URL}{FIREBASE_INSTALLATIONS_PATH}"

        response = requests.get(url, params=args)

        if not response:
            return ApiResponse(
                False, "No response from API in get_installation_by_id()"
            )

        if response.status_code != 200:
            return ApiResponse(
                False, None, f"get_installation_by_id() returned {response.status_code}"
            )

        reponse_json = response.json()

        if len(reponse_json) == 0 or installation_id not in reponse_json:
            return ApiResponse(False, None, "No Rointe installation found.")

        return ApiResponse(True, reponse_json[installation_id], None)

    def get_installations(self, local_id: str, auth_token: str) -> ApiResponse:
        """Retrieve the client's installations."""

        args = {"auth": auth_token, "orderBy": '"userid"', "equalTo": f'"{local_id}"'}
        url = f"{FIREBASE_DEFAULT_URL}{FIREBASE_INSTALLATIONS_PATH}"

        response = requests.get(url, params=args)

        if not response:
            return ApiResponse(False, "No response from API in get_installations()")

        if response.status_code != 200:
            return ApiResponse(
                False, None, f"get_installations() returned {response.status_code}"
            )

        reponse_json = response.json()

        if len(reponse_json) == 0:
            return ApiResponse(False, None, "No Rointe installations found.")

        installations = {}

        for key in reponse_json.keys():
            installations[key] = reponse_json[key]["location"]

        return ApiResponse(True, installations, None)

    def get_device(device_id: str, auth_token: str) -> ApiResponse:
        """Retrieve device data."""

        args = {"auth": auth_token}

        response = requests.get(
            "{}{}".format(
                FIREBASE_DEFAULT_URL, FIREBASE_DEVICES_PATH_BY_ID.format(device_id)
            ),
            params=args,
        )

        if not response:
            return ApiResponse(False, "No response from API in get_device()")

        if response.status_code != 200:
            return ApiResponse(
                False, None, f"get_device() returned {response.status_code}"
            )

        return ApiResponse(True, response.json(), None)

    def set_device_temp(
        self, device: RointeDevice, auth_token: str, new_temp: float
    ) -> bool:
        """Set the device target temperature."""

        device_id = device.id
        args = {"auth": auth_token}
        body = {"temp": new_temp, "mode": "manual"}

        url = "{}{}".format(
            FIREBASE_DEFAULT_URL, FIREBASE_DEVICE_DATA_PATH_BY_ID.format(device_id)
        )

        return self._send_patch_request(device_id, url, args, body)

    def set_device_preset(
        self, device: RointeDevice, auth_token: str, preset_mode: str
    ) -> bool:
        """Set the preset."""

        device_id = device.id
        args = {"auth": auth_token}
        body: Dict[str, Any] = {}

        url = "{}{}".format(
            FIREBASE_DEFAULT_URL, FIREBASE_DEVICE_DATA_PATH_BY_ID.format(device_id)
        )

        if preset_mode == "off":
            body = {"power": False, "mode": "manual"}
            return self._send_patch_request(device_id, url, args, body)

        elif preset_mode == "heat":
            body = {"mode": "manual", "power": True, "status": "none"}
            return self._send_patch_request(device_id, url, args, body)

        elif preset_mode == "auto":
            current_mode: ScheduleMode = device.get_current_schedule_mode()

            # For reasons unknown when changing modes we need to send the proper
            # temperature also.
            if current_mode == ScheduleMode.COMFORT:
                body = {"temp": device.comfort_temp}
            elif current_mode == ScheduleMode.ECO:
                body = {"temp": device.eco_temp}
            elif device.ice_mode:
                body = {"temp": device.ice_temp}
            else:
                body = {"temp": 20}

            request_power_status = self._send_patch_request(device_id, url, args, body)

            # and then set AUTO mode.
            request_mode_status = self._send_patch_request(
                device_id, url, args, {"mode": "auto", "power": True}
            )

            return request_power_status and request_mode_status

        else:
            _LOGGER.error("Invalid HVAC_MODE: %s", preset_mode)
            return False

    def _send_patch_request(
        self,
        device_id: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        body=None,
    ) -> bool:
        """Send a patch request."""

        body["last_sync_datetime_app"] = round(time.time() * 1000)

        _LOGGER.debug("Sending patch request body: %s", body)

        response = requests.patch(
            url,
            params=params,
            json=body,
        )

        if not response:
            _LOGGER.error(
                "No response from %s while setting sending %s to %s",
                AUTH_HOST,
                str(body),
                device_id,
            )
            return False

        if response.status_code != 200:
            _LOGGER.error(
                "Got response %s from %s while setting sending %s to %s",
                response.status_code,
                AUTH_HOST,
                str(body),
                device_id,
            )
            return False

        return True



"""Rointe app data model."""

from __future__ import annotations

from datetime import datetime
from enum import Enum

from rointesdk import utils


class DeviceMode(Enum):
    """Device working mode."""

    AUTO = "auto"
    MAN = "manual"


class ScheduleMode(Enum):
    """Schedule mode."""

    COMFORT = "C"
    ECO = "E"
    NONE = "O"


class RointeDevice:
    """Represent a Rointe device from the API."""

    id: str
    name: str
    serialnumber: str
    type: str
    product_version: str

    nominal_power: int
    power: bool

    # Describes the preset: ICE, ECO, Comfort, None
    preset: str
    mode: str

    temp: float
    temp_calc: float
    temp_probe: float

    # preset temperatures
    comfort_temp: float
    eco_temp: float
    ice_temp: float

    # User mode
    um_max_temp: float
    um_min_temp: float
    user_mode: bool
    ice_mode: bool

    # Schedule
    schedule: list[str]
    schedule_day: int
    schedule_hour: int

    last_sync_datetime_app: datetime
    last_sync_datetime_device: datetime

    def __init__(self, device_id: str, device_info: dict) -> None:
        """Initialize the device from the rointe's json blob."""
        self.id = device_id
        self.type = device_info["data"]["type"]
        self.product_version = str.lower(
            device_info["data"]["product_version"])
        self.serialnumber = device_info["serialnumber"]

        self.update_data(device_info)

    def update_data(self, device_info: dict) -> None:
        """Update the device data from a Json object."""

        data = device_info["data"]

        self.name = data["name"]
        self.nominal_power = int(data["nominal_power"])
        self.power = bool(data["power"])
        self.preset = data["status"]
        self.mode = data["mode"]

        self.temp = float(data["temp"])
        self.temp_calc = float(data["temp_calc"])
        self.temp_probe = float(data["temp_probe"])

        self.comfort_temp = float(data["comfort"])
        self.eco_temp = float(data["eco"])
        self.ice_temp = float(data["ice"])

        # User mode settings are only valid for V2 radiators.
        if self.product_version == "v2":
            self.um_max_temp = float(data["um_max_temp"])
            self.um_min_temp = float(data["um_min_temp"])
            self.user_mode = bool(data["user_mode"])
        else:
            self.user_mode = False

        self.ice_mode = bool(data["ice_mode"])
        self.schedule = data["schedule"]
        self.schedule_day = data["schedule_day"]
        self.schedule_hour = data["schedule_hour"]

        self.last_sync_datetime_app = datetime.fromtimestamp(
            int(data["last_sync_datetime_app"]) / 1000.0
        )

        self.last_sync_datetime_device = datetime.fromtimestamp(
            int(data["last_sync_datetime_device"]) / 1000.0
        )

    def get_current_schedule_mode(self) -> ScheduleMode:
        """Return the current schedule mode for the device.

        Returns C for Comfort, E for Eco, O for no-schedule
        """
        day_time = utils.now()
        day_of_week = day_time.weekday()  # 0 is Monday
        hour_index = day_time.hour

        current_mode = self.schedule[day_of_week][hour_index]

        if current_mode == "C":
            return ScheduleMode.COMFORT
        elif current_mode == "E":
            return ScheduleMode.ECO
        else:
            return ScheduleMode.NONE

    def user_mode_supported(self) -> bool:
        """Return True if this device supports user mode."""
        return self.product_version == "v2"
